#!/usr/bin/env python
"""
cli.py

Command line interface for tools in ms_thermo
"""

import click
from ms_thermo.tadia import tadia_table, tadia_cantera
from ms_thermo.yk_from_phi import yk_from_phi
from ms_thermo.fresh_gas import fresh_gas
from pkg_resources import resource_listdir, resource_string

@click.command()
@click.argument('temperature', nargs=1)
@click.argument('pressure', nargs=1)
@click.argument('phi', nargs=1)
def tadia_table_cli(temperature, pressure, phi):
	"""Calculate the adiabatic flame temperature of a kero-air mixture from 2S_KERO_BFER table"""
	burnt_temperature, yk = tadia_table(float(temperature), float(pressure), float(phi))
	print(f"\nThe adiabatic flame temperature of a mix C10H22-air from tables is : {burnt_temperature:.2f} K.")
	print("\nSpecies     |    Mass fraction")
	print("------------------------------")
	for specie in yk.keys():
		print(f'{specie:12s}|       {yk[specie]:.3f}')

@click.command()
@click.argument('temperature', nargs=1)
@click.argument('pressure', nargs=1)
@click.argument('phi', nargs=1)
@click.argument('fuel', nargs=1)
@click.argument('path2cti', nargs=1)
def tadia_cantera_cli(temperature, pressure, phi, fuel, path2cti):
	"""Calculate the adiabatic flame temperature of a fuel-air mixture from cantera"""
	burnt_temperature, yk = tadia_cantera(float(temperature), float(pressure), float(phi),
		str(fuel),str(path2cti))
	print(f"\nThe adiabatic flame temperature of a mix C10H22-air from cantera is : {burnt_temperature:.2f} K.")
	print("\nSpecies     |    Mass fraction")
	print("------------------------------")
	nb_other_species = 0
	y_other_species = 0
	for specie in yk.keys():
		if yk[specie] < 1e-3:
			nb_other_species +=1
			y_other_species += yk[specie]
		else:
			print(f'{specie:12s}|       {yk[specie]:.3f}')
	print(f'+ {nb_other_species} others |       {y_other_species:.3f}')

@click.command()
@click.argument('phi', nargs=1)
@click.argument('c_x', nargs=1)
@click.argument('h_y', nargs=1)
def yk_from_phi_cli(phi, c_x, h_y):
	"""Calculate the mass fractions of a fuel-air mixture"""
	yk = yk_from_phi(float(c_x), float(h_y), float(phi))
	print("\nSpecies     |    Mass fraction")
	print("------------------------------")
	for specie in yk.keys():
		print(f'{specie:12s}|       {yk[specie]:.3f}')


@click.command()
@click.argument('temperature', nargs=1, type=float)
@click.argument('pressure', nargs=1, type=float)
@click.argument('phi', nargs=1, type=float)
def fresh_gas_cli(temperature, pressure, phi):
	"""Calculate the conservative variables of a kero-air mixture"""
	rho,rhoE,rhoyk = fresh_gas(temperature, pressure, phi)
	print(f'\nrho       |  {rho:.3f} kg/m3')
	print(f'rhoE      |  {rhoE:.3f} J.kg/m3')
	print(f'rhoYk     |')            
	for specie in rhoyk.keys():
		print(f' {specie:9s}|  {rhoyk[specie]:.3f} mol.kg/m3')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# author: Benjamin FARCY, Anne FELDEN

""" Tool Flame param to get laminar premixed flame parameters such as
    laminar flame thickness and laminar flame speed from a table, with
    pressure, fresh gas temperature and equivalence ratio as entries.
"""
import numpy as np
import h5py
from scipy.interpolate import interpn
__all__ = ["FlameTable"]

class FlameTable():
    """ 
    *FlameTable class includes:*

    :param phi_list: list of equivalence ratios
    :param temperature_list: list of fresh gases temperatures
    :param pressure_list: list of pressures
    :param datanames_list: list of the stored data for the flames
    :param data_dict: dictionnary with all the stored data for the flames
    """

    def __init__(self):

        self.phi_list = None
        self.temperature_list = None
        self.pressure_list = None

        self.datanames_list = None
        self.data_dict = {}
        self.interpolated_data_dict = {}
        self.nparam = None

    def print_param_space(self):
        """ Pretty print param space """

        self.nparam = (len(self.phi_list) *
                       len(self.temperature_list) *
                       len(self.pressure_list))
        print(" >>> Parameters space")
        print(" >>> Number of operating conditions: " + str(self.nparam))
        print(str("eq ratio").ljust(20) + ": " + str(np.mean(self.phi_list))+ " [-]")
        print(str("temperature").ljust(20) + ": " + str(np.mean(self.temperature_list)) + " [K]")
        print(str("pressure").ljust(20) + ": " + str(np.mean(self.pressure_list)) + " [Pa]")

    def read_table_hdf(self, hdf_filename):
        """ 
        *Read the flame data table from an hdf format*
        
        :param hdf_filename: name of the input hdf file
        :param verbosity: display infos
        """

        hdf_file = h5py.File(hdf_filename, "r")

        self.phi_list = hdf_file["Data"]["EQUIVALENCE_RATIO"][:]
        self.temperature_list = hdf_file["Data"]["TEMPERATURE"][:]
        self.pressure_list = hdf_file["Data"]["PRESSURE"][:]


        self.datanames_list = [dataset for dataset in hdf_file["Data"].keys(
        ) if not "CLASS" in hdf_file["Data"][dataset].attrs.keys()]
        for dataset in self.datanames_list:
            self.data_dict[dataset] = hdf_file["Data"][dataset][:]

        dims = [hdf_file["Data"].attrs["DIM1_NAME"].decode(),
                hdf_file["Data"].attrs["DIM2_NAME"].decode(),
                hdf_file["Data"].attrs["DIM3_NAME"].decode()]


        if dims[0] != "EQUIVALENCE_RATIO":
            raise IOError("The first dimension of the table " +
                          "should be EQUIVALENCE_RATIO")
        if dims[1] != "TEMPERATURE":
            raise IOError("The second dimension of the table " +
                          "should be TEMPERATURE")
        if dims[2] != "PRESSURE":
            raise IOError("The third dimension of the table " +
                          "should be PRESSURE")

        hdf_file.close()

    def check_bounds(self, equivalence_ratio, temperature, pressure):
        """ Check that the input variables temperature, pressure and
            equivalence ratio are within the bounds of the stored data base
        """
        # print("\n ----", self.pressure_list)
        if (pressure < self.pressure_list[0]) or (pressure > self.pressure_list[-1]):
            raise IOError('Input pressure ' +
                          str(pressure) +
                          ' is out of database bounds: ' +
                          str(self.pressure_list[0]) + " " +
                          str(self.pressure_list[-1]))
        if (temperature < self.temperature_list[0]) or (temperature > self.temperature_list[-1]):
            raise IOError('Input temperature ' +
                          str(temperature) +
                          ' is out of database bounds: ' +
                          str(self.temperature_list[0]) + " " +
                          str(self.temperature_list[-1]))
        if (equivalence_ratio < self.phi_list[0]) or (equivalence_ratio > self.phi_list[-1]):
            raise IOError('Input equivalence_ratio ' +
                          str(equivalence_ratio) +
                          ' is out of database bounds: ' +
                          str(self.phi_list[0]) + " " +
                          str(self.phi_list[-1]))

    def get_params(self, equivalence_ratio=1.0, temperature=600, pressure=101325):
        """ Get the laminar flame params (flame speed, thickness and omega) from
            the pressure, the temperature and the equivalence ratio.
            It involves a trilinear interpolation in the table.
        """

        def squeeze_interp(in_index, in_data):
            """ Remove one data dimensions from the interpolation
                becaues interpn is not capable of handling it.
                Example:
                 in_index is a tuple:
                  in_index: ([0.5, 0.6], [300, 400], [101325])
                  out_index: ([0.5, 0.6], [300, 400])
                 in_data is an array:
                  in_data: [0.55, 350, 101325]
                  out_data: [0.55, 350]
            """
            out_index = ()
            out_data = []
            for i_elem, elem in enumerate(in_index):
                if len(elem) > 1:
                    out_index += (elem, )
                    out_data.append(in_data[i_elem])

            return out_index, out_data

        # bounds error is activated in interpn but it is more explicit
        # to do it externally
        self.check_bounds(equivalence_ratio, temperature, pressure)

        index, target_point = squeeze_interp((self.phi_list,
                                              self.temperature_list,
                                              self.pressure_list),
                                             [equivalence_ratio, temperature, pressure])

        for dataname in self.data_dict:
            self.interpolated_data_dict[dataname] = interpn(index,
                                                            np.squeeze(self.data_dict[dataname]),
                                                            target_point,
                                                            bounds_error=True)

    def print_interpolated_data(self):
        """ Print interpolated data """

        for dataname in self.interpolated_data_dict:
            print(str(dataname) + " = " + str(self.interpolated_data_dict[dataname]))



from ms_thermo.flame_params import FlameTable
from ms_thermo.yk_from_phi import yk_from_phi
import  pkg_resources

try:
	import cantera as ct
except ModuleNotFoundError:
	print('Cantera module is not installed. The function tadia_cantera will not work. Sorry about that..')

__all__ = ['tadia_cantera',
		   'tadia_table']
		   
def tadia_table(t_fresh_gases, p_fresh_gases, phi, fuel = None):
	"""
	*Compute the adiabatic flame temperature of a premixed kero/air mixture from tables*

	:param t_fresh_gases: Temperature of the fresh gases [K]
	:type t_fresh_gases: float
	:param p_fresh_gases: Pressure of the fresh gases [Pa]
	:type p_fresh_gases: float
	:param phi: Equivalence ratio [-]
	:type phi: float
	:param fuel: Fuel considered
	:type fuel: string

	:returns: 
		- **t_burnt_gases** - Temperature of the burnt gases
		- **yk_burnt** - Dict of mass fractions of burnt gases
	"""

	case = FlameTable()

	mass_molar = {
			"C": 0.0120107,
			"H": 0.00100797,
			"O2": 0.0319988,
			"N2": 0.0280134
			}

	if fuel == None:
		fuel_table = pkg_resources.resource_filename(
            __name__, f'../INPUT/2S_KERO_BFER.h5')

		c_x = 10
		h_y = 20
		yk_fresh = yk_from_phi(phi, c_x, h_y)
		
		mass_mol_fuel = (c_x * mass_molar["C"] + h_y * mass_molar["H"])
		coeff_o2 = c_x + (h_y / 4)
		stoechio = coeff_o2 * mass_molar["O2"] / mass_mol_fuel

		yk_burnt = dict()
		yk_burnt['N2'] = yk_fresh['N2']

		if phi<=1: 
			yk_burnt["fuel"] = 0
			yk_burnt["O2"] = yk_fresh["O2"] - stoechio * yk_fresh["fuel"]
			yk_burnt["CO2"] = c_x * yk_fresh["fuel"] * (mass_molar["O2"] + mass_molar["C"]) / mass_mol_fuel
			yk_burnt["H2O"] = h_y / 2 * yk_fresh["fuel"] * (0.5 * mass_molar["O2"] + 2 * mass_molar["H"]) / mass_mol_fuel
		else:
			yk_burnt["fuel"] = yk_fresh["fuel"] - yk_fresh["O2"] / stoechio
			yk_burnt["O2"] = 0
			yk_burnt["CO2"] = (1 / ( 1 + h_y / (4 * c_x))) * yk_fresh["O2"] * (mass_molar["O2"] + mass_molar["C"]) / mass_molar["O2"]
			yk_burnt["H2O"] = (1 / (1 / 2 + 2 * c_x / h_y)) * yk_fresh["O2"] * (0.5 * mass_molar["O2"] + 2 * mass_molar["H"]) / mass_molar["O2"]

	else:
		fuel_table = pkg_resources.resource_filename(
            __name__, f'../INPUT/'+fuel+'.h5')

	case.read_table_hdf(fuel_table)
	case.get_params(phi, t_fresh_gases, p_fresh_gases)
	
	t_burnt_gases = case.interpolated_data_dict['TEMPERATURE_BURNT'][0]

	return t_burnt_gases, yk_burnt

def tadia_cantera(t_fresh_gases, p_fresh_gases, phi, fuel='CH4', cti_file='gri30.cti'):
	"""
	*Compute the adiabatic flame temperature of a premixed fuel/air mixture from Cantera*

	:param t_fresh_gases: Temperature of the fresh gases [K]
	:type t_fresh_gases: float
	:param p_fresh_gases: Pressure of the fresh gases [Pa]
	:type p_fresh_gases: float
	:param phi: Equivalence ratio [-]
	:type phi: float
	:param fuel: Fuel considered
	:type fuel: string
	:param cti_file: Path to the cti file to consider
	:type cti_file: string

	:returns: 
		- **t_burnt_gases** - Temperature of the burnt gases
		- **yk_burnt** - Dict of mass fractions of burnt gases

	.. note:: **Warning**: This function may not be available if you do not have cantera in your environment
	"""
	case = ct.Solution(cti_file)
	case.TP=t_fresh_gases,p_fresh_gases
	case.set_equivalence_ratio(phi,{fuel:1},{'O2': 1, 'N2': 3.76})
	case.equilibrate('HP')
	t_burnt_gases = case.T
	yk_burnt = case.mass_fraction_dict()
	return t_burnt_gases, yk_burnt



"""

************
Examples for State.
************

State is an object handling the Internal state of a CFD fluid, namely:
- density
- Chemical Energy
- Mass Fractions

Limits of State object
======================
Velocity is NOT partof the state object and must be treated separately.
State refer to a point (shape 1) or a set of points (shape n).
The spatial aspects, i.e. the position of the points,
or the mesh, must be handled separately.
"""
import numpy as np
from ms_thermo import State#, build_thermo_from_cantera, build_thermo_from_avbp

def how_to_state_single():
    """

    live Example
    ============
    In the following a state for a single point is created
    using the primitive variables T, P, Y_k.
    """
    #Prerequisites : - a thermodynamic database (a default AVBP one is used)
    # species = build_thermo_from_cantera("./chem.cti")           #for cantera species database
    # species = build_thermo_from_chemkin("./thermo.dat")         #for chemkin species database
    # species = build_thermo_from_avbp("./species_database.dat")  #for avbp species database


    # State object initialization
    # basic : mass fractions, temperature, pressure setup
    # the variables can be in arrays of the same size
    state = State(temperature=600.,               # Kelvin, optional, deaults to 300.
                  pressure=100000.,               # Pa, optional, defaults to 101325.
                  mass_fractions_dict={'N2': 1.})#,#optional, defaults to {'O2':0.2325,'N2':0.7675}
                  # species_db=species)     # defaults to the avbp species_database.dat from pyavbp

    # this shows the most common value, the minima and maxima of the current state variables
    print(state)

    # reachable variables
    print(state.energy)         # in J
    print(state.rho)            # in Kg.m^-3
    print(state.mass_fracs)     # no unit
    print(state.temperature)    # in K
    print(state.pressure)       # in Pa

    # those can be updated directly from energy or rho:
    state.energy = 216038.00954532798
    state.rho = 1.171918080399414

    # or from primitives : temperature, pressure, mass_fractions:
    state.mass_fracs = {'O2': 0.2325, 'N2': 0.7675} #recomputes rho and mass fracs, keeps energy
    state.pressure = 101325.        #recomputes density, keeps mass fractions and energy
    state.temperature = 300.        #recomputes energy and density, keeps mass fractions

    #this can be useful to update the primitives all at once, or to ensure a meaningful state:
    state.update_state()

    #useful methods that does not update the state variables
    print(state.list_species())             # list of the species present in the mixture
    print(state.mix_energy(600.))           # gets the energy (J) value at given temperature (K)
    print(state.mix_enthalpy(600.))         # gets the enthalpy (J) value at given temperature (K)
    print(state.mix_molecular_weight())     # computes the total molecular weight of the mixture
    return state

def how_to_state_hundred():
    """

    live Example
    ============
    In the following a state for a hundred point is created
    using the primitive variables T, P, Y_k.
    """
    state = State(temperature=600. * np.ones(100),                  #optional
                  pressure=100000.* np.ones(100),                   #optional
                  mass_fractions_dict={'N2': 1. * np.ones(100)})    #optional
    #WARNING be careful with the variables shapes, the main ones are energy, rho and mass_fracs
    state.update_state(temperature=300. * np.ones(100),         #optional
                       pressure=101325.* np.ones(100),          #optional
                       mass_fracs={'O2': 1. * np.ones(100)})    #optional

    # this recomputes mass fractions (thus stores density and mass fractions)
    # then temperature (thus stores energy and density)
    # then pressure (thus stores density)
    return state

def how_to_state_conservatives():
    """

    live Example
    ============
    In the following a state for a single point is created
    using the conservative variables rho, rho_e, rho_y
    """
    # state can also be initalized from conservatives : rho, rho_e, rho_y in that order
    rho = 1.171918080399414
    rho_e = rho * 216038.00954532798
    rho_y = {'O2': 0.2325 * rho, 'N2': 0.7675 * rho}
    state_cons = State.from_cons(rho, rho_e, rho_y)
    print(state_cons.temperature)
    print(state_cons.energy)
    print(state_cons.mix_energy(300.))
    return state_cons

if __name__ == '__main__':
    _ = how_to_state_single()
    _ = how_to_state_hundred()
    _ = how_to_state_conservatives()



"""
This script calculate mass_fraction of species from a Phi
"""

__all__ = ["yk_from_phi",
           "phi_from_far"]

def phi_from_far(far, c_x, h_y):
    """
    *Return phi coefficient with the fuel air ratio coeff + fuel composition*

    :param far: the air-fuel ratio
    :type far: float
    :param c_x: stoechio coeff of Carbone
    :type c_x: float
    :param h_y: stoechio coeff of hydrogene
    :type h_y: float

    :returns:
        - **phi** - Equivalence ratio
    """
    mass_molar = {
        "C": 0.0120107,
        "H": 0.00100797,
        "O2": 0.0319988,
        "N2": 0.0280134
    }
    mass_mol_fuel = (c_x * mass_molar["C"] + h_y * mass_molar["H"])
    coeff_o2 = c_x + (h_y / 4)
    y_o2_fuel_sto = (coeff_o2 * mass_molar["O2"]) / mass_mol_fuel
    phi = far / y_o2_fuel_sto

    return phi

def yk_from_phi(phi, c_x, h_y):
    """ 
    *Return the mass fraction of species from a fuel aspect ratio and stoechio element coeff*

    :param phi: the air-fuel aspect ratio
    :type phi: float
    :param c_x: stoechio coeff of Carbone
    :type c_x: float
    :param h_y: stoechio coeff of hydrogene
    :type h_y: float

    :returns:
        - **y_k** - A dict of mass fractions

    """
    y_k = dict()
    mass_molar = {
        "C": 0.0120107,
        "H": 0.00100797,
        "O2": 0.0319988,
        "N2": 0.0280134
    }

    mass_mol_fuel = (c_x * mass_molar["C"] + h_y * mass_molar["H"])
    coeff_o2 = c_x + (h_y / 4)
    y_o2_fuel_sto = (coeff_o2 * mass_molar["O2"]) / mass_mol_fuel

    if phi == 0.:
        y_k["fuel"] = 0.
    else:
        y_k["fuel"] = (1. / (1. + (
            1.
            + 3.76 * (mass_molar["N2"] / mass_molar["O2"]))
            * (y_o2_fuel_sto / phi))
        )
    y_air = 1 - y_k["fuel"]
    y_k["N2"] = y_air / 1.303794
    y_k["O2"] = 0.303794 * y_k["N2"]

    return y_k
